#include <inttypes.h>
#include <xen/xen.h>

#include "emu64.h"

/* --------------------------------------------------------------------- */

static char *print_pgflags(uint32_t flags)
{
    static char buf[80];

    snprintf(buf, sizeof(buf), "%s%s%s%s%s%s%s%s%s\n",
	     flags & _PAGE_GLOBAL   ? " global"   : "",
	     flags & _PAGE_PSE      ? " pse"      : "",
	     flags & _PAGE_DIRTY    ? " dirty"    : "",
	     flags & _PAGE_ACCESSED ? " accessed" : "",
	     flags & _PAGE_PCD      ? " pcd"      : "",
	     flags & _PAGE_PWT      ? " pwt"      : "",
	     flags & _PAGE_USER     ? " user"     : "",
	     flags & _PAGE_RW       ? " write"    : "",
	     flags & _PAGE_PRESENT  ? " present"  : "");
    return buf;
}

void pgtable_walk(uint64_t va, uint64_t root_mfn)
{
    void *physmem = (void*)XEN_RAM_64;
    uint64_t *pgd, *pud, *pmd, *pte;
    uint64_t mfn;
    uint32_t slot, flags;

    if (0 == root_mfn)
	root_mfn = addr_to_frame(read_cr3());

    printk("page table walk for va %" PRIx64 ", root_mfn %" PRIx64 "\n",
	   va, root_mfn);

    pgd   = physmem + frame_to_addr(root_mfn);
    slot  = PGD_INDEX_64(va);
    mfn   = get_pgframe_64(pgd[slot]);
    flags = get_pgflags_64(pgd[slot]);
    printk("pgd   : %p +%3d  |  mfn %4" PRIx64 "  |  %s",
	   pgd, slot, mfn, print_pgflags(flags));
    if (!(flags & _PAGE_PRESENT))
	return;

    pud   = physmem + frame_to_addr(mfn);
    slot  = PUD_INDEX_64(va);
    mfn   = get_pgframe_64(pud[slot]);
    flags = get_pgflags_64(pud[slot]);
    printk(" pud  : %p +%3d  |  mfn %4" PRIx64 "  |  %s",
	   pud, slot, mfn, print_pgflags(flags));
    if (!(flags & _PAGE_PRESENT))
	return;

    pmd   = physmem + frame_to_addr(mfn);
    slot  = PMD_INDEX_64(va);
    mfn   = get_pgframe_64(pmd[slot]);
    flags = get_pgflags_64(pmd[slot]);
    printk("  pmd : %p +%3d  |  mfn %4" PRIx64 "  |  %s",
	   pmd, slot, mfn, print_pgflags(flags));
    if (!(flags & _PAGE_PRESENT))
	return;
    if (flags & _PAGE_PSE)
	return;

    pte   = physmem + frame_to_addr(mfn);
    slot  = PTE_INDEX_64(va);
    mfn   = get_pgframe_64(pte[slot]);
    flags = get_pgflags_64(pte[slot]);
    printk("   pte: %p +%3d  |  mfn %4" PRIx64 "  |  %s",
	   pte, slot, mfn, print_pgflags(flags));
}

int pgtable_fixup_flag(uint64_t va, uint32_t flag)
{
    void *physmem = (void*)XEN_RAM_64;
    uint64_t *pgd, *pud, *pmd, *pte;
    uint32_t slot;
    int fixes = 0;

    /* quick test on the leaf page via linear page table */
    pte = find_pte_64(va);
    if (!test_pgflag_64(*pte, flag)) {
	*pte |= flag;
	return 1;
    }

    /* do full page table walk */
    pgd   = physmem + read_cr3();
    slot  = PGD_INDEX_64(va);
    if (!test_pgflag_64(pgd[slot], flag)) {
	pgd[slot] |= flag;
	fixes++;
    }

    pud   = physmem + frame_to_addr(get_pgframe_64(pgd[slot]));
    slot  = PUD_INDEX_64(va);
    if (!test_pgflag_64(pud[slot], flag)) {
	pud[slot] |= flag;
	fixes++;
    }

    pmd   = physmem + frame_to_addr(get_pgframe_64(pud[slot]));
    slot  = PMD_INDEX_64(va);
    if (!test_pgflag_64(pmd[slot], flag)) {
	pmd[slot] |= flag;
	fixes++;
    }
    return fixes;
}

/* --------------------------------------------------------------------- */

void *map_64(uint64_t maddr)
{
    void *ram = (void*)XEN_RAM_64;
    return ram + maddr;
}

uint64_t *find_pte_64(uint64_t va)
{
    uint64_t *lpt_base = (void*)XEN_LPT_64;
    uint64_t offset = (va & 0xffffffffffff) >> PAGE_SHIFT;

    return lpt_base + offset;
}

void update_emu_mappings_64(uint64_t cr3_mfn)
{
    uint64_t *old_pgd;
    uint64_t *new_pgd;
    int idx;

    new_pgd  = map_64(frame_to_addr(cr3_mfn));

    idx = PGD_INDEX_64(XEN_M2P_64);
    if (!test_pgflag_64(new_pgd[idx], _PAGE_PRESENT)) {
	/* new one, must init static mappings, copy from current */
//	printk("%s: %lx is new, init\n", __FUNCTION__, cr3_mfn);
	old_pgd  = map_64(read_cr3() & PAGE_MASK);
	while (idx < PGD_INDEX_64(XEN_DOM_64)) {
	    new_pgd[idx] = old_pgd[idx];
	    idx++;
	}
    }

    /* linear pgtable mapping */
    idx = PGD_INDEX_64(XEN_LPT_64);
    new_pgd[idx] = get_pgentry_64(cr3_mfn, LPT_PGFLAGS);
}

